import os 
import logging
from typing import Generator
from .walker import BaseModuleWalker, ModuleWalker


logger = logging.getLogger(__name__)

class DiscoverEngine:
    def __init__(self, file_path: str, module_walker_type: BaseModuleWalker,
                 ignore_syntax_errors = True):
        if not os.path.exists(file_path):
            logger.error(f'{file_path!r} does not exists')
            raise FileNotFoundError(f'{file_path!r} does not exist')
        
        self.file_path = file_path 
        self._package = os.path.split(file_path)[-1]
        self.walker = module_walker_type 
        self.ignore_syntax_error = ignore_syntax_errors

    @property
    def package(self):
        return self._package

    def __iter__(self):
        for dirpath, folders, files in os.walk(self.file_path):
            logger.debug(f"Directory Path: {dirpath!r}")
            
            if dirpath.endswith("__pycache__"):
                continue 
            for file in files:
                if not file.endswith(".py"):
                    logger.debug(f"{os.path.join(dirpath,file)!r} is not a python file.")
                    continue
                
                full_path = os.path.join(dirpath,file)
                logger.info(f'{full_path!r} will be processed.')

                module_name = os.path.splitext(full_path)[0]\
                                .replace(self.file_path,self.package)\
                                .replace(os.sep,".").strip(" .")
                logger.info(F'Module Path: {module_name}')
                logger.debug('Initializing debug.')

                m = self.walker(module_name = module_name, module_path = full_path)
                logger.debug(f'Walker initialzed')
                
                try:
                    dct = m()
                except SyntaxError as e:
                    if not self.ignore_syntax_error:
                        raise e
                    logger.exception(f'Error when trying to parse module')
                    dct = {}
                
                module_name = module_name.replace("__init__","").strip(' .')
                yield {module_name:dct}


import os
import logging 
from datetime import datetime
from .walker import ModuleWalker 
from .engine import DiscoverEngine as Discover 
from .exporters import (
    JSONExporter, 
    YamlExporter,
    SqliteExporter,
    StreamExporter
)



logger = logging.getLogger(__name__)

# Add a handler to the package logger so subloggers propagate to it
handler = logging.FileHandler(os.path.join(os.getcwd(),f'pypkgview_{int(datetime.now().timestamp())}.log'),mode = 'w')
handler.setFormatter(logging.Formatter("[%(asctime)s]: [%(module)s]: [%(levelname)s]: [%(funcName)s]: [%(lineno)d]: [%(message)s]"))
handler.setLevel(logging.DEBUG)
logger.addHandler(handler)

logger.info('pypkgview initialized.')


import argparse
import logging  
from . import (ModuleWalker, 
               Discover,JSONExporter, 
               YamlExporter,
               SqliteExporter, StreamExporter)
from .datastructures import Exporter
from .render import extractor


parser = argparse.ArgumentParser("pypkgview",description = "Understand any Python package before you touch it.")

subparser = parser.add_subparsers(dest = "command",required = True)

exporter = subparser.add_parser("export",
    help="Export package analysis to json, yaml or sqlite",
    description="Walk a Python package and export extracted analysis data.")
exporter.add_argument('--module-path','-m', required = True)
exporter.add_argument('--exporter','-e',choices = ["json","yaml","sqlite"])
exporter.add_argument('--verbose','-v',action='store_true')

dashboard = subparser.add_parser("dashboard",
    help="Generate an HTML dashboard from package analysis",
    description="Analyze a package and render an interactive HTML dashboard."
)
dashboard.add_argument('--module-path','-m', required = True)
dashboard.add_argument('--verbose','-v',action='store_true')


args = parser.parse_args()

if args.verbose:
    logging.getLogger().setLevel(logging.DEBUG)

logger = logging.getLogger(__name__)


logger.debug(f'{parser = !r}')

discover = Discover(file_path = args.module_path, module_walker_type = ModuleWalker)

logger.info('Discover initalized.')

logger.info(f'Command initialized: {args.command!r}')
match args.command:
    case "export":
        print(f"Exporter \033[1;38;5;9mExporting to {args.exporter!r}.\033[0m")
        exporter: Exporter = {
            "json": JSONExporter, 
            "yaml": YamlExporter, 
            "sqlite": SqliteExporter
        }.get(args.exporter,StreamExporter)

        logger.info(f'Exporter: {exporter.__class__.__name__!r} initalized')
        exporter().export(discover = discover)
        print(f"\033[1;38;5;9mExporting to {args.exporter!r} is complete\033[0m")
    
    case "dashboard":
        print(f"Dashboard: \033[1;38;5;9mExporting data to sqlite exporter.\033[0m")
        SqliteExporter().export(discover = discover)
        extractor(discover.package)
    case _:
        print(f"\033[38;5;11m[WARNING] Select a valid option\033[0m")





import ast
import logging
from .datastructures import Class 

logger = logging.getLogger(__name__)

class NodeVisitor(ast.NodeVisitor):
    def __init__(self):
        super().__init__()
        self.classes: list[Class] = []
        self._classes_stack: list[Class] = [] 
        self.functions: list[ast.FunctionDef] = []
        self.imports: list[ast.Import] = []
        self.import_froms: list[ast.ImportFrom] = []
        self.variables: list[ast.Assign] = []
        logger.debug('NodeVisitor initalized.')

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        logger.debug('visit_ClassDef called.')
        
        cls = Class(name = node.name, bases = node.bases, keywords = node.keywords, decorator_list = node.decorator_list,
                is_nested_class = bool(self._current_class))
        logger.debug(f'{cls} initalized.')

        logger.debug(f'{self._current_class}')
        if self._current_class is not None:
            cls.parent_class = self._current_class.name
        self.classes.append(cls)
        self._classes_stack.append(cls)
        self.generic_visit(node)
        self._classes_stack.pop()

    @property
    def _current_class(self) -> Class | None:
        logger.debug(f'{self._classes_stack = !r}')
        return self._classes_stack[-1] if self._classes_stack else None
        

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        logger.debug('visit_FunctionDef called.')

        is_method = self._current_class is not None 
        logger.debug(f'{node.name!r} is method {is_method!r}')
        
        if is_method:
            method_type = list(filter(
                lambda x: x in ["staticmethod", "property", "classmethod", "setter", "deleter"],
                map(lambda x: getattr(x,{ast.Name:"id", ast.Attribute: "attr"}.get(x.__class__)) if not isinstance(x,ast.Call) else "", 
                    node.decorator_list)
                ))
            
            logger.debug(f'{method_type = !r}')

            match method_type:
                
                case []:
                    self._current_class.method.append(node)
                case ["staticmethod"]:
                    self._current_class.staticmethod.append(node)
                case ["classmethod"]:
                    self._current_class.classmethod.append(node)
                case ["property" | "setter" | "deleter"]:
                    self._current_class.property.append(node)
                
            if node.name in ["__get__","__set__", "__delete__"]:
                self._current_class.is_descriptor = True 
                logger.debug(f'{node.name!r} is a descriptor True')
                match node.name:
                    case "__set__" | "__delete__":
                        self._current_class.descriptor_type = "data"
                    case "__get__":
                        if self._current_class.descriptor_type is None: 
                            self._current_class.descriptor_type = "non_data"


        else:
            self.functions.append(node)
        
    visit_AsyncFunctionDef = visit_FunctionDef
    def visit_Import(self, node: ast.Import) -> None:
        self.imports.append(node)
        self.generic_visit(node)

    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        self.import_froms.append(node)
        self.generic_visit(node)

    def visit_Assign(self, node: ast.Assign) -> None:
        if self._current_class is None:
            self.variables.append(node)


import sys
import json
import logging
import sqlite3
import pathlib 
from jinja2 import Environment, FileSystemLoader 


logger = logging.getLogger(__name__)

def classify_imports(package_name,imports):
    profile = {"stdlib": 0,"internal": 0,"third_party":0}
    stdlib    = sys.stdlib_module_names 

    for imp in imports:
        name = imp["name"]
        count = imp["count"]

        if name == package_name:
            profile["internal"] += count 
        
        elif name in stdlib:
            profile["stdlib"] += count
        else:
            profile["third_party"] += count 

    return profile 
        


def get_insights(package_name:str,totals: dict,import_details: list[dict[str,str]]) -> list[dict]:
    insights = []

    # class vs function ratio
    classes   = totals["classes"]
    functions = totals["functions"]
    ratio     = classes / functions if functions > 0 else 0

    if ratio > 2:
        insights.append({
            "title":  "OOP Heavy",
            "detail": f"{classes:,} classes vs {functions:,} functions — "
                      f"this package strongly favors class-based design.",
        })
    elif ratio < 0.5:
        insights.append({
            "title":  "Functional API",
            "detail": f"{functions:,} functions vs {classes:,} classes — "
                      f"this package is primarily a functional interface.",
        })

    import_profile = classify_imports(package_name,import_details)
    stdlib      = import_profile["stdlib"]
    internal    = import_profile["internal"]
    third_party = import_profile["third_party"]
    total       = stdlib + internal + third_party

    if total > 0:
        tp_pct  = round(third_party / total * 100)
        int_pct = round(internal    / total * 100)
        sb_pct  = round(stdlib      / total * 100)

        if tp_pct > 40:
            insights.append({
                "title":  "High External Dependency",
                "detail": f"{tp_pct}% of imports are third-party packages. "
                          f"This package relies heavily on external libraries."
            })
        elif int_pct > 60:
            insights.append({
                "title":  "Self-Contained",
                "detail": f"{int_pct}% of imports are internal. "
                          f"This package mostly depends on itself."
            })
        elif sb_pct > 50:
            insights.append({
                "title":  "Stdlib Reliant",
                "detail": f"{sb_pct}% of imports come from the standard library. "
                          f"Minimal external dependencies."
            })
    

    

    return insights,import_profile



def render_dashboard(data: dict, output_path: str):
    
    env = Environment(
        loader = FileSystemLoader(pathlib.Path(__file__).parent/'dashboard/templates')
    )

    chartjs = pathlib.Path(pathlib.Path(__file__).parent/"dashboard/static/chart.min.js").read_text()
    
    template = env.get_template("dashboard.html")
    insights,import_profile = get_insights(data["package_name"],data["totals"],data["first_imp_names"])
    html = template.render(
        chartjs_source = chartjs,
        package_name = data["package_name"],
        totals = data["totals"],
        insights = insights,
        top_external_imports=data["top_external_imports"],  
        top_internal_imports=data["top_internal_imports"],  
        top_decorators=data["top_decorators"],
        top_bases=data["top_bases"],
        top_modules=data["top_modules"],
        class_properties=data["class_properties"], 
        import_profile = import_profile,
    func_properties=data["func_properties"],  
    )

    pathlib.Path(output_path).write_text(html,encoding = 'utf-8')
    print("Dashboard: \033[38;5;10mCompleted Dashboard printed\033[0m")

def extractor(package_name: str):
    queries: dict[str,str] = json.load(open(pathlib.Path(__file__).parent/"dashboard/query_lib.json"))
    
    results = {"package_name": package_name}
    logger.debug('Establishing connection to sqlite')
    conn = sqlite3.connect(f'{package_name}.sqlite')
    cursor = conn.cursor()

    print(f"Dashboard: \033[38;5;9mFetching Package Details\033[0m")
    logger.debug('Fetching Package Details')

    m,c,f,cnst,imps = cursor.execute(queries["Package Composition"]).fetchone()
    results["totals"] = {"modules":m, "classes":c,"functions":f,
                         "globals": cnst,"imports":imps}
    

    print(f"Dashboard: \033[38;5;9mFetching Class Details\033[0m")
    logger.debug('Fetching Class Details')
    
    dec,ctx,nes, met, itb, itr, dcpr = cursor.execute(queries["Class Details"]).fetchone()
    results["class_properties"] = {"decorated":dec, 
                                   "contextmanager":ctx, "nested":nes, 
                                   "has_metaclass": met, 
                                   "iterable": itb, "iterator": itr,
                                   "descriptor": dcpr}
    
    print(f"Dashboard: \033[38;5;9mFetching Functions Details\033[0m")
    logger.debug('Fetching Functions Details')

    dec, gen, asyc, gen_deg = cursor.execute(queries["Functions Details"]).fetchone()
    results["func_properties"] = {
        "decorated": dec, 
        "generator": gen,
        "async": asyc, 
        "generator_deleg": gen_deg
    }

    print("Dashboard: \033[38;5;9mFetching Top External Imports\033[0m")
    logger.debug('Fetching Top External Imports')
    cursor.execute(queries["Top External Imports"],{'mod':f'{package_name}%'})

    results["top_external_imports"] = []
    for fname, count in cursor:
        results["top_external_imports"].append(
            {"name": fname, "count": count}
        )

    print("Dashboard: \033[38;5;9mFetching Top Internal Imports\033[0m")
    logger.debug('Fetching Top Internal Imports')
    cursor.execute(queries["Top Internal Imports"],{'mod':f'{package_name}%'})


    results["top_internal_imports"] = []
    for fname, count in cursor:
        results["top_internal_imports"].append(
            {"name": fname, "count": count}
        )

    print("Dashboard: \033[38;5;9mFetching Top Inherited Classes\033[0m")
    logger.debug('Fetching Top Inherited Classes')
    cursor.execute(queries["Top Inherited Classes"])

    results["top_bases"] = []
    for fname, count in cursor:
        results["top_bases"].append(
            {"name": fname, "count": count}
        )

    print("Dashboard: \033[38;5;9mFetching Top Decorators\033[0m")
    logger.debug('Fetching Top Decorators')
    cursor.execute(queries["Top Decorators"])

    results["top_decorators"] = []
    for fname, count in cursor:
        results["top_decorators"].append(
            {"name": fname, "count": count}
        )

    print("Dashboard: \033[38;5;9mFetching Modules Detail\033[0m")
    logger.debug('Fetching Modules Detail')
    cursor.execute(queries["god_modules"])

    results["top_modules"] = []
    for name, c, f,g,imp,total in cursor:
        results["top_modules"].append(
            {
                "name": name,
                "classes": c, "functions": f, 
                "globals":g, "imports":imp, "total":total
            }
        )





    logger.debug('Fetching first name of imports')
    cursor.execute(queries["_Import Split"],{'mod':f'{package_name}%'})
    results["first_imp_names"] = []

    for name,count in cursor:
        results["first_imp_names"].append(
            {"name": name, "count":count}
        )

    print(f'Dashboard: \033[1;38;5;10mComplete Fetching\033[0m')
    
    render_dashboard(results,f'{package_name}.html')



    



    

    



    


    


    

    





    


# render_dashboard(
#     {"package_name": "torch",
#      "totals": {
#          "modules": 2113,
#          "classes": 4886,
#          "functions": 15030,
#          "globals": 10956,
#          "imports": 29620
#      },
#      "top_external_imports": [
#     {"name": "typing.Any",              "count": 720},
#     {"name": "typing.Optional",         "count": 525},
#     {"name": "collections.abc.Callable","count": 510},
#     {"name": "typing.Union",            "count": 366},
#     {"name": "logging",                 "count": 326},
#     {"name": "typing.TYPE_CHECKING",    "count": 294},
#     {"name": "functools",               "count": 291},
#     {"name": "collections.abc.Sequence","count": 238},
#     {"name": "os",                      "count": 234},
#     {"name": "__future__.annotations",  "count": 222},
# ],
# "top_internal_imports": [
#     {"name": "torch._C",                "count": 891},
#     {"name": "torch.nn.functional",     "count": 743},
#     {"name": "torch._tensor",           "count": 651},
#     {"name": "torch.autograd",          "count": 589},
#     {"name": "torch._utils",            "count": 498},
#     {"name": "torch.nn.modules",        "count": 445},
#     {"name": "torch._ops",              "count": 401},
#     {"name": "torch.fx",                "count": 387},
#     {"name": "torch._prims",            "count": 334},
#     {"name": "torch.distributed",       "count": 298},
#     ],
#     "top_bases": [
#     {"name": "nn.Module",           "count": 341},
#     {"name": "enum.Enum",           "count": 141},
#     {"name": "abc.ABC",             "count": 86},
#     {"name": "typing.NamedTuple",   "count": 77},
#     {"name": "autograd.Function",   "count": 71},
#     {"name": "builtins.Exception",  "count": 66},
#     {"name": "VariableTracker",     "count": 65},
#     {"name": "typing.Generic",      "count": 56},
#     {"name": "typing.Protocol",     "count": 40},
#     {"name": "HigherOrderOperator", "count": 37},
# ],
# "top_decorators": [
#     {"name": "dataclasses.dataclass",     "count": 738},
#     {"name": "out_wrapper",               "count": 331},
#     {"name": "register_decomposition",    "count": 330},
#     {"name": "_onnx_symbolic",            "count": 319},
#     {"name": "contextlib.contextmanager", "count": 259},
#     {"name": "register_meta",             "count": 248},
#     {"name": "parse_args",                "count": 188},
#     {"name": "register_lowering",         "count": 177},
#     {"name": "functools.cache",           "count": 145},
#     {"name": "fx.compatibility",          "count": 133},
# ],
# "top_bases": [
#     {"name": "nn.Module",            "count": 341},
#     {"name": "enum.Enum",            "count": 141},
#     {"name": "abc.ABC",              "count": 86},
#     {"name": "typing.NamedTuple",    "count": 77},
#     {"name": "autograd.Function",    "count": 71},
#     {"name": "builtins.Exception",   "count": 66},
#     {"name": "VariableTracker",      "count": 65},
#     {"name": "typing.Generic",       "count": 56},
#     {"name": "typing.Protocol",      "count": 40},
#     {"name": "HigherOrderOperator",  "count": 37},
# ],
# "top_decorators": [
#     {"name": "dataclasses.dataclass",      "count": 738},
#     {"name": "out_wrapper",                "count": 331},
#     {"name": "register_decomposition",     "count": 330},
#     {"name": "_onnx_symbolic",             "count": 319},
#     {"name": "contextlib.contextmanager",  "count": 259},
#     {"name": "register_meta",              "count": 248},
#     {"name": "parse_args",                 "count": 188},
#     {"name": "register_lowering",          "count": 177},
#     {"name": "functools.cache",            "count": 145},
#     {"name": "fx.compatibility",           "count": 133},
# ],
# "top_modules": [
#     {"name": "common_methods_invocations", "classes": 10,  "functions": 468, "globals": 19,  "imports": 168, "total": 665},
#     {"name": "_sfdp_pattern_16",           "classes": 0,   "functions": 0,   "globals": 504, "imports": 20,  "total": 524},
#     {"name": "torch._refs",                "classes": 0,   "functions": 292, "globals": 117, "imports": 61,  "total": 470},
#     {"name": "_inductor.lowering",         "classes": 0,   "functions": 270, "globals": 78,  "imports": 105, "total": 453},
#     {"name": "common_utils",               "classes": 34,  "functions": 163, "globals": 98,  "imports": 87,  "total": 382},
#     {"name": "_meta_registrations",        "classes": 1,   "functions": 315, "globals": 16,  "imports": 44,  "total": 376},
#     {"name": "_dynamo.utils",              "classes": 20,  "functions": 191, "globals": 53,  "imports": 109, "total": 373},
#     {"name": "_inductor.ir",               "classes": 96,  "functions": 36,  "globals": 16,  "imports": 128, "total": 276},
#     {"name": "decompositions",             "classes": 1,   "functions": 221, "globals": 6,   "imports": 38,  "total": 266},
#     {"name": "torch (root)",               "classes": 24,  "functions": 53,  "globals": 32,  "imports": 145, "total": 254},
# ],
#     "class_properties": {
#     "decorated":       1097,
#     "contextmanager":   127,
#     "nested":            57,
#     "has_metaclass":     28,
#     "iterable":          85,
#     "iterator":           8,
#     "descriptor":        10,
# },
# "func_properties": {
#     "decorated":        740,
#     "generator":         48,
#     "async":              3,
#     "generator_deleg":   18,
# },

#     }, 
#     "dashboard.html"   
# )


import os 
import logging
from typing import Protocol
from .datastructures import Discover as EngineProtocol


logger = logging.getLogger(__name__)

class StreamExporter:
    
    def  export(self, discover: EngineProtocol): 
        logger.info(f'StreamExporter chosen.')

        import pprint
        print(f"Package: \033[1;38;5;9m{discover.package}\033[0m")
        for dct in discover:
            module_name = list(dct)[0]
            print(f"Parsing: \033[1;38;5;10m{discover.package}\033[0m")
            pprint.pprint(dct)




class YamlExporter:
    def  export(self, discover: EngineProtocol): 
        
        logger.info(f'YamlExporter chosen.')
        try:
            from collections import defaultdict
            import yaml
        except ModuleNotFoundError:
            logger.exception("error when trying to load module yaml")
            raise 
                
        yaml.SafeDumper.add_representer(defaultdict,yaml.representer.Representer.represent_dict)
        cwd = os.getcwd()
        with open(os.path.join(cwd, f'{discover.package}.yaml'), 'w', encoding = 'utf-8') as file:
            for dct in discover:
                module_name = list(dct)[0]
                print(f'Parsing: \033[1;38;5;10m{module_name!r}\033[0m')
                yaml.safe_dump(dct, file)
        
    

class JSONExporter:
    def  export(self, discover: EngineProtocol): 
        logger.info(f'JSONExporter chosen.')
        import json
        
        
        dirpath = os.path.join(os.getcwd(), f'{discover.package}')
        os.makedirs(dirpath, exist_ok = True)
        for dct in discover:
            module_name, *_ = dct
            print(f'Parsing: \033[1;38;5;10m{module_name!r}\033[0m')
            fdr = ".".join(module_name.split(".")[:2])
            os.makedirs(os.path.join(dirpath, fdr), exist_ok=True)
            with open(os.path.join(dirpath, fdr, f'{module_name}.json'), 'w', encoding = 'utf-8') as f:
                json.dump(dct,f)

         
class SqliteExporter:
    def export(self, discover: EngineProtocol):
        logger.info(f'Sqlitexporter chosen.')
        import sqlite3 

        logger.debug("creating table")
        cwd = os.getcwd()
        conn = sqlite3.connect(os.path.join(cwd, f'{discover.package}.sqlite'))
        cursor = conn.cursor()

        cursor.executescript("""
        DROP TABLE IF EXISTS modules;
        CREATE TABLE modules(
            id INTEGER PRIMARY KEY,
            name TEXT);

        DROP TABLE IF EXISTS classes;
        CREATE TABLE classes (
            id INTEGER PRIMARY KEY,
            module_id INTEGER REFERENCES modules(id),
            name TEXT,
            is_descriptor INTEGER,
            descriptor_type TEXT, -- non_data, data
            is_nested INTEGER,
            is_contextmanager INTEGER,
            is_iterable INTEGER,
            is_iterator INTEGER,
            parent_class TEXT,
            has_metaclass INTEGER,
            metaclass TEXT);

        DROP TABLE IF EXISTS bases;
        CREATE TABLE bases (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            class_id INTEGER REFERENCES classes(id),
            name TEXT);

        DROP TABLE IF EXISTS functions;
        CREATE TABLE functions (
            id INTEGER PRIMARY KEY ,
            module_id INTEGER REFERENCES modules(id),
            name TEXT,
            is_decorated INTEGER,
            is_async INTEGER,
            is_generator INTEGER,
            has_generator_delegation INTEGER);

        DROP TABLE IF EXISTS decorators;
        CREATE TABLE decorators (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            class_id INTEGER REFERENCES classes(id),
            function_id INTEGER REFERENCES functions(id),
            name TEXT);

        DROP TABLE IF EXISTS constants;                     
        CREATE TABLE constants (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            module_id INTEGER REFERENCES modules(id),
            name TEXT,
            type TEXT -- 'constant' | 'variable'
        );
                             
        DROP TABLE IF EXISTS imports;
        CREATE TABLE imports (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            module_id INTEGER REFERENCES modules(id),
            source TEXT,   -- null for direct imports
            name TEXT,
            alias TEXT,
            type  TEXT-- 'direct' | 'external' | 'internal_absolute' | 'internal_relative'
        );
        CREATE INDEX idx_imports_name ON imports(name);
        CREATE INDEX idx_functions_name ON functions(name);
        CREATE INDEX idx_classes_name ON classes(name);
        CREATE INDEX idx_functions_module ON functions(module_id);
        CREATE INDEX idx_classes_module ON classes(module_id);
        """)
        logger.debug('Finished creating table')
        class_id = 0 
        func_id = 0
        for idx, dct in enumerate(discover, start = 1):
            module_name = list(dct)[0]
            print(f'Parsing: \033[1;38;5;10m{module_name!r}\033[0m')
            cursor.execute("INSERT INTO modules VALUES(?,?)", (idx,module_name))
            class_insert_stmt = """INSERT INTO classes (id, module_id, name, is_descriptor, 
                               descriptor_type, is_nested, metaclass, parent_class, has_metaclass,
                               is_contextmanager,is_iterable,is_iterator)
                               VALUES 
                               (:id, :module_id, :name, :is_descriptor, :descriptor_type,
                               :is_nested, :metaclass,
                               :parent_class, :has_metaclass,:is_contextmanager,
                               :is_iterable, :is_iterator)"""
            bases_stmt = """INSERT INTO bases(class_id, name) VALUES (:class_id, :name)"""
            decorator_stmt = """INSERT INTO decorators(class_id, function_id, name) VALUES 
                            (:class_id, :function_id, :name)"""
            functions_stmt = """INSERT INTO functions 
                (id, module_id, name, is_async, is_generator, has_generator_delegation, is_decorated) VALUES
                (:id, :module_id, :name, :is_async, :is_generator, :has_generator_delegation, :is_decorated)"""
            
            for class_name, cls in dct[module_name]["classes"].items():
                class_id += 1
                cursor.execute(class_insert_stmt,{"id": class_id, "module_id": idx, 
                    "name": class_name,
                    "is_descriptor": int(cls["is_descriptor"]), 
                    "descriptor_type": cls.get("descriptor_type"),
                    "is_nested": int(cls["is_nested"]), 
                    "parent_class": cls["parent_class"],
                    "has_metaclass": int(cls["metadata"]["has_metaclass"]),
                    "is_contextmanager": int(cls["is_contextmanager"]),
                    "is_iterable": int(cls["is_iterable"]),
                    "is_iterator": int(cls["is_iterator"]),
                    "metaclass": cls["metadata"]["attrs"].get("metaclass")})
                cursor.executemany(bases_stmt,[
                               {"class_id": class_id, "name": base}
                               for id, base in enumerate(cls["bases"], start = 1)])
                cursor.executemany(decorator_stmt,[
                               {"class_id": class_id, "function_id": None, "name": dec}
                               for id, dec in enumerate(cls["decorators"], start = 1)])
            
            for func_name, func in dct[module_name]["functions"].items():
                
                func_id += 1
                cursor.execute(functions_stmt,
                    {"id": func_id, "module_id": idx, "name": func_name, 
                     "is_async": int(func["is_async"]),
                     "is_generator": int(func["is_generator"]), 
                     'is_decorated': int(func['is_decorated']),
                     "has_generator_delegation": int(func["has_generator_delegation"])})
                
                cursor.executemany(decorator_stmt, 
                               [{"class_id": None, "function_id": func_id, "name": dec}
                                   for dec in func["decorators"]
                               ])
            cursor.executemany("""INSERT INTO constants(module_id,name,type)
                               VALUES(:module_id, :name, :type)""",
                               [{"module_id": idx, "name": cnst,
                                 "type": {"constants":"constant","variables": "variable"}[cnst_type]}
                                   for cnst_type, consts in dct[module_name]["constants"].items()
                                   for cnst in consts
                               ])
            
            import_stmt = """INSERT INTO imports(module_id, source, name, alias,type)
            VALUES(:module_id, :source, :name, :alias,:type)"""
            
            
            cursor.executemany(import_stmt,[
                {"module_id": idx, "source": None, 
                "name": (m_name := x.split(" as "))[0].strip(), 
                "alias": m_name[-1].strip() if len(m_name) > 1  else None,
                "type": "direct"}
                for x in dct[module_name]["imports"]["direct"]
            ])

            cursor.executemany(import_stmt,[
                {"module_id": idx, "source": src, 
                 "name": (m_name := name.split(" as "))[0].strip(), 
                "alias": m_name[-1].strip() if len(m_name) > 1  else None, 
                 "type": "internal_absolute" if src == discover.package else "external"
                }
                for src, names in dct[module_name]["imports"]["external_imports"].items()
                for name in names
            ])
            
            cursor.executemany(import_stmt,[
                {"module_id": idx, "source": src, 
                 "name": (m_name := name.split(" as "))[0].strip(), 
                "alias": m_name[-1].strip() if len(m_name) > 1  else None, 
                 "type": {"absolute_imports": "internal_absolute",
                          "relative_imports": "internal_relative"}[type]
                }
                for type, imp_dct in dct[module_name]["imports"][discover.package].items()
                for src, names in imp_dct.items()
                for name in names
            ])
        cursor.close()
        conn.commit()
        conn.close()
                

